/* Created by Richard Wade
 *
 * This is a work in progress for controlling a self-balancing robot
 * using a MPU6050 Gyroscope/Accelerometer module and an Arduino.
 * It uses a Kalman filter for the signals from the MPU and PID to 
 * calculate the speed and direction that the motors should turn in
 * order to maintain balance. Motor selection and play in the gears will
 * affect the robots ablility to stay balanced
 * 
 */


#include <Wire.h>
#include <SPI.h>  
#include "I2Cdev.h"
#include "MPU6050.h"

#include <LiquidCrystal_I2C.h>//this is the special I2C LCD display library that came with the display

LiquidCrystal_I2C myDisplay(0x27,16,2);  //this instantiates an LCD object named "myDisplay"
                                         //We set the LCD address to 0x27 (this is a hexadecimal number, which is equal to 0b100111); this is the I2C 
                                         //bus address for the PCF8574 port extender chip that controls the display (see tutorial video)
                                         //and define the number of columns (16) and rows (2) of the display
                                         //this tells the methods in the library what display we are dealing with

MPU6050 accelgyro;
MPU6050 initialize;
int16_t ax, ay, az;
int16_t gx, gy, gz;

#define Gry_offset 0  //The offset of the gyro
#define Gyr_Gain 16.348
#define Angle_offset 5.8  // The offset of the accelerator
#define RMotor_offset 35  // The offset of the Motor
#define LMotor_offset 34  // The offset of the Motor
#define pi 3.14159 

float kp, ki, kd;
float Angle_Raw, Angle_Filtered, omega;
float Turn_Speed = 0, Turn_Speed_K = 0;
float Run_Speed = 0, Run_Speed_K = 0;
float LOutput,ROutput;

unsigned long preTime = 0;
unsigned long lastTime;
float Input, Output;
float errSum, dErr, error, lastErr;
int timeChange; 

int TN1=4;
int TN2=3;
int ENA=9;
int TN3=6;
int TN4=5;
int ENB=10;

struct Axis  // Datas from remote control 
{
  uint16_t axis_1;
  uint16_t axis_2;
  uint16_t axis_3;
  uint16_t axis_4;
  uint16_t axis_5;
  uint16_t axis_6;
  uint16_t axis_7;
  uint16_t axis_8;
};
Axis axis_x;

struct Gesture  // Data from MPU 
{
  float angle;
  float omega;
  int speed; 
  int P;
  int I;
  int D;
  uint16_t null_1; 
  uint16_t null_2;
};
Gesture data;

void setup()
{
  Wire.begin();
  Serial.begin(115200);

  myDisplay.init();                      //initialize the lcd - this sets the character canvas to 5x8 pixels and some other hardware specifics
                                         //Note: This .init() method also starts the I2C bus, i.e. there does not need to be
                                         //a separate "Wire.begin();" statement in the setup.
 
  myDisplay.backlight();//this turns the backlight on
  //Print a message to the LCD.  

  /* If the robot was turned on with the angle over 45(-45) degrees,the wheels will not spin until the robot is in right position. */
  accelgyro.initialize();
  initialize.setFullScaleGyroRange(MPU6050_GYRO_FS_2000);;
  for(int i = 0; i < 200; i++)  // Looping 200 times to get the real gesture when starting
  {
    Filter();
  }
  if (abs(Angle_Filtered) < 45)  // Start the robot after cleaning data
  {
    omega = Angle_Raw = Angle_Filtered = 0;
    Filter();
    Output = error = errSum = dErr = Run_Speed = Turn_Speed = 0;
    myPID();
  }

  pinMode(TN1,OUTPUT);
  pinMode(TN2,OUTPUT);
  pinMode(TN3,OUTPUT);
  pinMode(TN4,OUTPUT);
  pinMode(ENA,OUTPUT);
  pinMode(ENB,OUTPUT);
  
}


void loop() 
{
  Filter();
  Serial.print("  Angle = ");
  Serial.print(Angle_Filtered);
  
  myDisplay.setCursor(5,1);
  myDisplay.print(" Angle");myDisplay.print(Angle_Filtered);

  // If angle > 45 or < -45 then stop the robot
  if (abs(Angle_Filtered) < 45)
  {
    //Recive();
    myPID();
    PWMControl();
  }
  else
  {
    analogWrite(ENA, 0);  // Stop the wheels
    analogWrite(ENB, 0);  // Stop the wheels
    
    for(int i = 0; i < 50; i++)  // Keep reading the MPU gesture after falling down
    {
      Filter();
    }
    
    if(abs(Angle_Filtered) < 45)  // Empty data and restart the robot automaticly
    {
      for(int i = 0; i <= 100; i++)  // Reset the robot and delay 2 seconds before turning motors
      {
        omega = Angle_Raw = Angle_Filtered = 0;
        Filter();
        Output = error = errSum = dErr = Run_Speed = Turn_Speed = 0;
        myPID();
      }
    }
  }
  Serial.println();
}

// Kalman filtering for the MPU6050 signal to increase accuracy
void Filter()
{
  // Raw datas from MPU6050
  accelgyro.getMotion6(&ax, &ay, &az, &gx, &gy, &gz);
  Angle_Raw = (atan2(ay, az) * 180 / pi + Angle_offset);
  omega =  gx / Gyr_Gain + Gry_offset;
  // Filters data to get the real gesture
  unsigned long now = millis();
  float dt = (now - preTime) / 1000.00;
  preTime = now;
  float K = 0.8;
  float A = K / (K + dt);
  Angle_Filtered = A * (Angle_Filtered + omega * dt) + (1 - A) * Angle_Raw;
  // Preparing data which will to be shown on the LCD.
  data.omega = omega;
  data.angle = Angle_Filtered;
}

void myPID()
{ 
  // You can see the values both on the serial monitor and the LCD screen.
  kp = analogRead(A0) * 0.01;      //Serial.print("  kp = "); Serial.print(kp);
  ki = analogRead(A1) * 0.0001;   //Serial.print("  ki = "); Serial.print(ki);
  kd = analogRead(A2) * 1.0;      //Serial.print("  kd = "); Serial.print(kd);
  // Preparing datas which need to be shown on the LCD.
  data.P = analogRead(A0);
  data.I = analogRead(A1);
  data.D = analogRead(A2);
  
  //then print at the appropriate cursor position myDisplay.print("  kp=");myDisplay.print(kp);
  myDisplay.setCursor(0,0);
  myDisplay.print("kp");myDisplay.print(kp);  
  myDisplay.setCursor(8,0);
  myDisplay.print(" ki");myDisplay.print(ki);
  myDisplay.setCursor(0,1);
  myDisplay.print("kd");myDisplay.print(kd);
 
  // Calculating the output values using the gesture values and the PID values.
  unsigned long now = millis();
  timeChange = (now - lastTime);
  lastTime = now;
  
  error = Angle_Filtered;  // Proportion
  errSum += error * timeChange;  // Integration
  dErr = (error - lastErr) / timeChange;  // Differentiation
  Output = kp * error + ki * errSum + kd * dErr;
  lastErr = error;
 
 // Turn_Speed and Run_Speed inicialized at zero but will be changed by a bluetooth function
 // to be added once balnce is acheived
  LOutput = Output - Turn_Speed + Run_Speed;
  ROutput = Output + Turn_Speed + Run_Speed;
  data.speed = (LOutput + ROutput) / 2;
}


// This function controls the motor speed and direction using pulse width modulation
void PWMControl()
{
  if(LOutput > 0)
  {
    digitalWrite(TN1, HIGH);
    digitalWrite(TN2, LOW);
  }
  else if(LOutput < 0)
  {
    digitalWrite(TN1, LOW);
    digitalWrite(TN2, HIGH);
  }
  else
  {
    digitalWrite(ENA, 0);
  }
  if(ROutput > 0)
  {
    digitalWrite(TN3, HIGH);
    digitalWrite(TN4, LOW);
  }
  else if(ROutput < 0)
  {   
    digitalWrite(TN3, LOW);
    digitalWrite(TN4, HIGH);
  }
  else
  {
    digitalWrite(ENB, 0);
  }
    analogWrite(ENA, min(255, (abs(LOutput) + LMotor_offset)));
    analogWrite(ENB, min(255, (abs(ROutput) + RMotor_offset)));
}
